--
-- FillVolume - HardPoint
--
-- @author:    	Xentro (Marcus@Xentro.se)
-- @website:	www.Xentro.se
-- @history:	v1.0 - 2015-10-18 - Initial implementation
-- 

FillVolume = {};

function FillVolume:prerequisitesPresent(specializations)
	return true;
end;

-- pre check before attaching
function FillVolume:allowAdding(vehicle)
	if vehicle.fillLevel > 0 then
		local txt;
		
		if g_i18n:hasText("ERROR_WE_ARE_LOADED") then
			txt = g_i18n:getText("ERROR_WE_ARE_LOADED");
		end;
		
		return false, txt;
	end;
	
	return true;
end;

function FillVolume:allowRemoval(vehicle)
	-- we do the same check as allowAdding...
	local allow, txt = FillVolume.allowAdding(nil, vehicle);
	
	return allow, txt;
end;

function FillVolume:load(xmlFile, parentClassObj)
	local sv = self.vehicle;
	self.bodyData = {};
	
	local capacityNotice = false;
	local capacity = getXMLFloat(xmlFile, "vehicle.capacity");
	if capacity ~= nil then
		self.bodyData["capacity"] = sv:getCapacity();
		sv:setCapacity(capacity);
		capacityNotice = true;
	else
		capacity = sv:getCapacity();
	end;
	
    local fillTypes = getXMLString(xmlFile, "vehicle.fillTypes#fillTypes");
    if fillTypes ~= nil then
		self.bodyData["fillTypes"] = sv.fillTypes;
		sv.fillTypes = {};
		sv.fillTypes[Fillable.FILLTYPE_UNKNOWN] = true;
		
        local types = Utils.splitString(" ", fillTypes);
        for k, v in pairs(types) do
            local fillType = Fillable.fillTypeNameToInt[v];
            if fillType ~= nil then
                sv.fillTypes[fillType] = true;
            else
                print("Warning: '" .. self.configFileName .. "' has invalid fillType '" .. v .. "'.");
            end;
        end;
    end;
	
	if self.isClient then
        local fillVolumes = {};
        local fillVolumeDeformers = {};
		
        local i = 0;
        while true do
            local key = string.format("vehicle.fillVolumes.fillVolume(%d)", i);
            if not hasXMLProperty(xmlFile, key) then break;	end;
            local fillVolume = {};

            fillVolume.baseNode = Utils.indexToObject(self.components, getXMLString(xmlFile, key .. "#index"));
            fillVolume.allSidePlanes = Utils.getNoNil(getXMLBool(xmlFile, key .. "#allSidePlanes"), true);
			local defaultFillType = getXMLString(xmlFile, key .. "#defaultFillType");
            
			if defaultFillType ~= nil then
                local fillType = Fillable.fillTypeNameToInt[defaultFillType];
                
				if fillType ~= nil then
					fillVolume.defaultFillType = fillType;
                else
					print("Warning: Invalid defaultFillType '" .. tostring(defaultFillType) .. "' in '" .. self.configFileName .. "'");
                end;
            end;

            fillVolume.maxDelta = Utils.getNoNil(getXMLFloat(xmlFile, key .. "#maxDelta"), 1.0);
            fillVolume.maxSurfaceAngle = math.rad( Utils.getNoNil(getXMLFloat(xmlFile, key .. "#maxAllowedHeapAngle"), 35) );

            local maxPhysicalSurfaceAngle = math.rad(35);
            fillVolume.maxSubDivEdgeLength = Utils.getNoNil(getXMLFloat(xmlFile, key .. "#maxSubDivEdgeLength"), 0.9);

            fillVolume.volume = createFillPlaneShape(fillVolume.baseNode, "fillPlane", capacity, fillVolume.maxDelta, fillVolume.maxSurfaceAngle, maxPhysicalSurfaceAngle, fillVolume.maxSubDivEdgeLength, fillVolume.allSidePlanes);

            fillVolume.deformers = {};
            if fillVolume.volume ~= nil then
                local j = 0;
                while true do
                    local node = Utils.indexToObject(self.components, getXMLString(xmlFile, key .. ".deformNode(" .. j .. ")#index"));
                    if node == nil then break; end;
					
                    local initPos = {localToLocal(node, fillVolume.baseNode, 0, 0, 0)};
                    local polyline = findPolyline(fillVolume.volume, initPos[1], initPos[3]);
                    fillVolumeDeformers[node] = {node = node, initPos = initPos, posX = initPos[1], posZ = initPos[3], polyline = polyline, volume = fillVolume.volume, baseNode = fillVolume.baseNode};
                    j = j + 1;
                end;
            end;

            fillVolume.scrollSpeedDischarge = {Utils.getVectorFromString(Utils.getNoNil(getXMLString(xmlFile, key .. "#scrollSpeedDischarge"), "0 0 0"))};
            fillVolume.scrollSpeedLoad = {Utils.getVectorFromString(Utils.getNoNil(getXMLString(xmlFile, key .. "#scrollSpeedLoad"), "0 0 0"))};
            for i = 1, 3 do
                fillVolume.scrollSpeedDischarge[i] = fillVolume.scrollSpeedDischarge[i] / 1000;
                fillVolume.scrollSpeedLoad[i] = fillVolume.scrollSpeedLoad[i] / 1000;
            end;
            fillVolume.uvPosition = {0, 0, 0};

            if fillVolume.volume ~= nil and fillVolume.volume ~= 0 then
                link(fillVolume.baseNode, fillVolume.volume);
                table.insert(fillVolumes, fillVolume);
            end;

            i = i + 1;
        end;

        fillVolumeHeights = {};
        fillVolumeHeightRefNodeToFillVolumeHeight = {};
        local i = 0;
        while true do
            local key = string.format("vehicle.fillVolumeHeights.fillVolumeHeight(%d)", i);
            if not hasXMLProperty(xmlFile, key) then break; end;
			
            local volumeHeight = {};
            volumeHeight.fillVolumeId = getXMLInt(xmlFile, key .. "#fillVolumeId");
            volumeHeight.volumeHeightIsDirty = false;

            volumeHeight.refNodes = {};
            local j = 0;
            while true do
                local refNode = Utils.indexToObject(self.components, getXMLString(xmlFile, string.format("%s.refNode(%d)#index", key, j)));
                if refNode == nil then break; end;
				
                table.insert(volumeHeight.refNodes, {refNode = refNode});
                fillVolumeHeightRefNodeToFillVolumeHeight[refNode] = volumeHeight;
                j = j + 1;
            end;

            volumeHeight.nodes = {};
            local j = 0;
            while true do
                local node = Utils.indexToObject(self.components, getXMLString(xmlFile, string.format("%s.node(%d)#index", key, j)));
				
                if node ~= nil then
                    local baseScale = {Utils.getVectorFromString(Utils.getNoNil(getXMLString(xmlFile, string.format("%s.node(%d)#baseScale", key, j)), "1 1 1"))};
                    local scaleAxis = {Utils.getVectorFromString(Utils.getNoNil(getXMLString(xmlFile, string.format("%s.node(%d)#scaleAxis", key, j)), "0 0 0"))};
                    local scaleMax = {Utils.getVectorFromString(Utils.getNoNil(getXMLString(xmlFile, string.format("%s.node(%d)#scaleMax", key, j)), "0 0 0"))};
                    local basePosition = {getTranslation(node)};
                    local transAxis = {Utils.getVectorFromString(Utils.getNoNil(getXMLString(xmlFile, string.format("%s.node(%d)#transAxis", key, j)), "0 0 0"))};
                    local transMax = {Utils.getVectorFromString(Utils.getNoNil(getXMLString(xmlFile, string.format("%s.node(%d)#transMax", key, j)), "0 0 0"))};
                    local orientateToWorldY = Utils.getNoNil(getXMLBool(xmlFile, string.format("%s.node(%d)#orientateToWorldY", key, j)), false);
                    table.insert(volumeHeight.nodes, {node = node, baseScale = baseScale, scaleAxis = scaleAxis, scaleMax = scaleMax, basePosition = basePosition, transAxis = transAxis, transMax = transMax, orientateToWorldY = orientateToWorldY});
				else
					break;
				end;
				
                j = j + 1;
            end;

            table.insert(fillVolumeHeights, volumeHeight);
            i = i + 1;
		end;
		
		if table.getn(fillVolumes) > 0 then
			self.bodyData["fillVolumes"] = sv.fillVolumes;
			self.bodyData["fillVolumeDeformers"] = sv.fillVolumeDeformers;
			
			sv.fillVolumes = fillVolumes;
			sv.fillVolumeDeformers = fillVolumeDeformers;
			
			for _, fillVolume in pairs(self.bodyData["fillVolumes"]) do
				if fillVolume.volume ~= nil then
					setVisibility(fillVolume.volume, false);
				end;
			end;
			
			for _, fillVolume in pairs(sv.fillVolumes) do
				if fillVolume.volume ~= nil then
					setVisibility(fillVolume.volume, sv.fillLevel > 0);
					
					if sv.fillLevel > 0 then
						local materialId, isDefault = Fillable.getFillMaterial(sv.currentFillType, Fillable.MATERIAL_TYPE_FILLPLANE, 1);
						
						if isDefault and fillVolume.defaultFillType ~= nil then
							materialId = Fillable.getFillMaterial(fillVolume.defaultFillType, Fillable.MATERIAL_TYPE_FILLPLANE, 1);
						end;
						
						if materialId ~= nil then
							setMaterial(fillVolume.volume, materialId, 0);
						end;
					end;
				end;
			end;
			
			local fillSourceStruct = {x = 0, y = 0, z = 0, d1x = 0, d1y = 0, d1z = 0, d2x = 0, d2y = 0, d2z = 0};
			sv:setFillLevel(0, sv.currentFillType, true, fillSourceStruct);
		else
			if capacityNotice then
				print("HP_FillVolume - Warning: Capacity have changed but no new fillVolume has been defined, old fillVolume wont fill passed " .. self.bodyData["capacity"] .. " (new capacity: " .. capacity .. ")");
			end;
		end;
		
		if table.getn(fillVolumeHeights) > 0 then
			self.bodyData["fillVolumeHeights"] = sv.fillVolumeHeights
			self.bodyData["fillVolumeHeightRefNodeToFillVolumeHeight"] = sv.fillVolumeHeightRefNodeToFillVolumeHeight;
			
			sv.fillVolumeHeights = fillVolumeHeights;
			sv.fillVolumeHeightRefNodeToFillVolumeHeight = fillVolumeHeightRefNodeToFillVolumeHeight;
		end;
	end;
	
	local exactFillRootNode = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.exactFillRootNode#index"));
	if exactFillRootNode ~= nil then
		self.bodyData["exactFillRootNode"] = sv.exactFillRootNode;
		g_currentMission.nodeToVehicle[sv.exactFillRootNode] = nil;
		sv.exactFillRootNode = exactFillRootNode;
		g_currentMission.nodeToVehicle[exactFillRootNode] = sv;
	end;
	
	local fillAutoAimTargetNode = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.fillAutoAimTargetNode#index"));
	if fillAutoAimTargetNode ~= nil then
		self.bodyData["fillAutoAimTargetNode"] = sv.fillAutoAimTargetNode;
		sv.fillAutoAimTargetNode = fillAutoAimTargetNode;
	end;
end;

function FillVolume:delete()
	local sv = self.vehicle;
	
	if self.bodyData["capacity"] ~= nil then
		sv:setCapacity(self.bodyData["capacity"]);
	end;
	
	if self.bodyData["fillTypes"] ~= nil then
		sv.fillTypes = self.bodyData["fillTypes"];
	end;
	
	if self.bodyData["fillVolumes"] ~= nil then
		local fillVolumes = sv.fillVolumes;
		
		sv.fillVolumes = self.bodyData["fillVolumes"];
		sv.fillVolumeDeformers = self.bodyData["fillVolumeDeformers"];
		
		for _, fillVolume in pairs(fillVolumes) do
			if fillVolume.volume ~= nil then
				delete(fillVolume.volume);
			end;
			
			fillVolume.volume = nil;
		end;
		
		local fillSourceStruct = {x = 0, y = 0, z = 0, d1x = 0, d1y = 0, d1z = 0, d2x = 0, d2y = 0, d2z = 0};
		self.vehicle:setFillLevel(self.vehicle.fillLevel, self.vehicle.currentFillType, false, fillSourceStruct);
	end;
	
	if self.bodyData["fillVolumeHeights"] ~= nil then
		sv.fillVolumeHeights = self.bodyData["fillVolumeHeights"];
		sv.fillVolumeHeightRefNodeToFillVolumeHeight = self.bodyData["fillVolumeHeightRefNodeToFillVolumeHeight"];
	end;
	
	if self.bodyData["exactFillRootNode"] ~= nil then
		g_currentMission.nodeToVehicle[sv.exactFillRootNode] = nil;
		
		sv.exactFillRootNode = self.bodyData["exactFillRootNode"];
		g_currentMission.nodeToVehicle[sv.exactFillRootNode] = sv;
	end;
	
	if self.bodyData["fillAutoAimTargetNode"] ~= nil then
		sv.fillAutoAimTargetNode = self.bodyData["fillAutoAimTargetNode"];
	end;
end;

-- depending on how your trailer is set for fillTypes we might need to save and load seperate to default
function FillVolume:loadAttributesAndNodes(xmlFile, key, resetVehicles)
	local fillLevel = getXMLFloat(xmlFile, key .. "#fillLevel");
	local fillType = getXMLString(xmlFile, key .. "#fillType");
	
	if fillLevel ~= nil and fillType ~= nil then
		local fillTypeInt = Fillable.fillTypeNameToInt[fillType];
		
		if fillTypeInt ~= nil then
			local fillSourceStruct = {x = 0, y = 0, z = 0, d1x = 0, d1y = 0, d1z = 0, d2x = 0, d2y = 0, d2z = 0};
			self.vehicle:setFillLevel(fillLevel, fillTypeInt, true, fillSourceStruct);
		end;
	end;
end;	

function FillVolume:getSaveAttributesAndNodes(nodeIdent)
	if self.bodyData["fillTypes"] ~= nil then
		local fillType = Fillable.fillTypeIntToName[self.vehicle.currentFillType];
		if fillType == nil then fillType = "unknown"; end;
		
		local attributes = 'fillLevel="' .. self.vehicle.fillLevel .. '" fillType="' .. fillType .. '"';
		
		return attributes;
	else
		return;
	end;
end;

function FillVolume:readStream(streamId, connection)
end;

function FillVolume:writeStream(streamId, connection)
end;

function FillVolume:update(dt)
end;

function FillVolume:updateTick(dt)
end;

function FillVolume:draw()
end;